#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <iostream>
#include <vector>

#include "socklib.hpp"
#include "readerwriter.hpp"
#include "bufferedreaderwriter.hpp"

typedef void (*sighandler_t)(int);

#define MESS_RIEN 0
#define MESS_FIN 1
#define MESS_ENTIER 2
#define MESS_TEXT 3

#define TAILLEBUFF 100

int  quesepassetil;

void traite_signal(int s) {
  // Cette fonction met à jour une variable globale
  // en fonction du signal reçu.
  // La variable sera lue après la fonction pause() dans le code du père
  if (s == SIGCHLD) {
    quesepassetil = MESS_FIN;
  } else if (s == SIGUSR1) {
    quesepassetil = MESS_ENTIER;
  } else if (s == SIGUSR2) {
    quesepassetil = MESS_TEXT;
  }
}


void fonction_fils(int nbiteration, int entreepipe) {
  int i;
  int r;
  int pere = getppid();
  socklib::BufferedReaderWriter out(entreepipe);

  srand(time(NULL));

  for (i=0; i<nbiteration; i++) {
    int val = rand()%2;
    
    usleep(rand()%3000000);

    if (val) {
      fprintf(stdout, "fils envoie %d\n", i);

      // J'utilise l'écriturre de bas niveau
      // si j'en utilise une autre (via les string ou les vector),
      // il faut que je crée un vector de taille 4 et que je le remplisse 
      out.write((char *)&i, 0, sizeof(int));

      
      // envoie le signal correspondant au père
      r = kill(pere, SIGUSR1);
      if (r) {
	perror("kill");
	exit(1);
      }
    } else {
      // Construction d'un message de taille 100
      // les 100 cases sont remplies de 0 car ça ne coûte
      // rien et c'est toujours mieu
      std::vector<char> mess(TAILLEBUFF, 0); 
      strcpy(mess.data(), "coucou");


      // je peux utiliser '<<' car j'ai remplie le tableau mess par des 0.
      std::cout << "Envoie du du message " << mess.data() << std::endl; 
      out.write(mess);

      // envoie le signal correspondant au père
      r = kill(pere, SIGUSR2);
      if (r) {
	perror("kill");
 	exit(1);
      }
    }
  }
}


int main(int argc, char *argv[]) {
  int tabpipe[2];
  int r;
  int nbiter;


  // lecture des arguments.
  // s'il y en a 1 c'est le nombre de message,
  // sinon, ce nombre est 10.
  if (argc == 1) {
    nbiter = 10;
  } else if ((argc == 2)&&(atoi(argv[1])>0)) {
    nbiter = atoi(argv[1]);
  } else {
    fprintf(stderr, "usage : %s <nb iterations>\n", argv[0]);
    exit(1);
 }

  // Création du pipe de discution
  r = pipe(tabpipe);
  if (r == -1) {
    perror("pipe");
    exit(1);
  }

  // lancement du fils
  pid_t pid = fork();
  if (pid == -1) {
    perror("fork()");
    exit(1);
  }

  if (pid == 0) {
    // code du fils
    // fermeture de la valeur non utilisée
    close(tabpipe[0]);
    // J'appelle une fonction
    // pour simplifier le code
    fonction_fils(nbiter, tabpipe[1]);
    // fin du fils
    close(tabpipe[1]);
    // ce close est inutile car la fermeture est faite
    // lors de la destruction du BufferedReaderWtriter
    exit(0);
  }

  // code du père

  // fermeture de la partie non utilisée du pipe
  close(tabpipe[1]);

  // construction de l'objet de 
  socklib::BufferedReaderWriter in(tabpipe[0]);

  // gestion du signal du fils
  sighandler_t sh;

  // cas du premier signal
  sh = signal(SIGUSR1, traite_signal);
  if (sh == SIG_ERR) {
    perror("signal");
    exit(1);
  }
  // cas du second signal (on appel la même fonction,
  // c'est elle qui fera la différence
  sh = signal(SIGUSR2, traite_signal);
  if (sh == SIG_ERR) {
    perror("signal");
    exit(1);
  }
  // Cas de la fin du fils
  sh = signal(SIGCHLD, traite_signal);
  if (sh == SIG_ERR) {
    perror("signal");
    exit(1);
  }

  quesepassetil = MESS_RIEN;
  while(1) {
    // on attend un signal
    pause();
    
    if (quesepassetil == MESS_FIN) {
      // fin du fils, on sort de la boucle
      break;
    } else if (quesepassetil == MESS_ENTIER) {
      int val;

      // lecture d'un entier (je lit dans un tableau
      // de la taille d'un entier, puis je recopie les bits
      // au bon endroit
      std::vector<char> res = in.read_all(sizeof(int));
      memcpy(&val, res.data(), sizeof(int));

      std::cout << "Recu une valeur entiere : " << val <<  std::endl;

    } else if (quesepassetil == MESS_TEXT) {

      // lecture du tableau de taille 100
      std::vector<char> res = in.read_all(TAILLEBUFF);
      // Je recopie le tableau dans une string qui sera plus simple à afficher. 
      std::string mess;
      mess.assign(res.data(), TAILLEBUFF);

      std::cout << "Recu un message text    : " << mess << std::endl;
    } else {
      // quesepassetil a une valeur non définie
      fprintf(stderr, "cela ne devrait pas arriver\n");
      exit(1);
    }

    quesepassetil = MESS_RIEN;
  }

  
  int status;
  pid_t res;

  // enterrement propre du fils (j'ai recopié le code
  // du tp shell).
  res = waitpid(pid, &status, 0);
  if (res < 0) {
    fprintf(stderr, "erreur avec le waitpid %s\n", strerror(errno));
    exit(1);
  } else if (WIFEXITED(status)) {
    fprintf(stdout, "le fils s'est terminée en renvoyant %d\n", WEXITSTATUS(status));
  } else if (WIFSIGNALED(status)) {
    fprintf(stdout, "le fils s'est terminée suite au signal %d\n", WTERMSIG(status));
  } else if (WIFSTOPPED(status)) {
    fprintf(stdout, "le fils à été stoppée par le signal %d\n", WSTOPSIG(status));
  } else if (WIFCONTINUED(status)) {
    fprintf(stdout, "le fils à été relancée\n");
  } else {
    fprintf(stderr, "cela ne devrait pas arriver\n");
  }
  
  return 0;
}


